/*****************************************************************/
/*                                                               */
/*  Exact arithmetics based on double precision floats ...       */
/*                                                               */
/*  Leif Kobbelt                              06/12/96           */
/*                                                               */
/*  Computer Graphics Group                                      */
/*  University of Erlangen-N"urnberg                             */
/*  Am Weichselgarten 9                                          */
/*  91058 Erlangen, GERMANY                                      */
/*                                                               */
/*  kobbelt@informatik.uni-erlangen.de                           */
/*                                                               */
/*  No commercial use without consent from the author! Please    */
/*  mind the copyright and give credits if you use the code      */
/*  within some software package.                                */
/*                                                               */
/*****************************************************************/

#include "ex_arith_double.hh"

/*****************************************************************/

#define  IEEE_double_Accu_Size  4096   //  = 2 * exponent range

#define  SPLIT_27  134217729.0         //  = (2^27 + 1)

#define  exponent(x)  ((*((int*) &(x)    ) & 0x7FF00000) >> 19)
#define  parity(x)     (*((int*) &(x) + 1) & 0x00000001)

/*****************************************************************/

IEEE_double_Accu::IEEE_double_Accu()
  {
    int i;

    t = new double[IEEE_double_Accu_Size];

    for (i=0 ; i<IEEE_double_Accu_Size ; t[i++] = 0.0);
  }

/*****************************************************************/

IEEE_double_Accu::IEEE_double_Accu(IEEE_double_Accu& A)
  {
    int i;

    t = new double[IEEE_double_Accu_Size];

    for (i=0 ; i<IEEE_double_Accu_Size ; t[i++] = A.t[i]);
  }

/*****************************************************************/

IEEE_double_Accu::~IEEE_double_Accu()
  {
    delete[] t;
  }

/*****************************************************************/

void IEEE_double_Accu::push(double x)
  {
    int i = parity(x) + exponent(x);

    while (t[i] != 0.0)
      {
        x    = x + t[i];
        t[i] = 0.0;
        i    = parity(x) + exponent(x);
      }

    t[i] = x;
  }

/*****************************************************************/

void IEEE_double_Accu::push(double x,double y)
  {
    double a,x1,x2,y1,y2;

    a  = SPLIT_27 * x;
    x1 = a - (a - x);
    x2 = x - x1;
    a  = SPLIT_27 * y;
    y1 = a - (a - y);
    y2 = y - y1;

    push(x1*y1);
    push(x1*y2);
    push(x2*y1);
    push(x2*y2);
  }

/*****************************************************************/

void IEEE_double_Accu::add(IEEE_double_Accu& A)
  {
    int i;

    for (i=0 ; i<IEEE_double_Accu_Size ; i++)
      if (A.t[i] != 0.0)
        push(A.t[i]);
  }

/*****************************************************************/

void IEEE_double_Accu::scale(double m)
  {
    int i;

    IEEE_double_Accu A;

    for (i=0 ; i<IEEE_double_Accu_Size ; i++)
      if (t[i] != 0.0)
        A.push(m,t[i]);

    double *tt = t;    
    t          = A.t;
    A.t        = tt;
  }

/*****************************************************************/

void IEEE_double_Accu::mult(IEEE_double_Accu& A)
  {
    int i;

    IEEE_double_Accu B;

    for (i=0 ; i<IEEE_double_Accu_Size ; i++)
      if (A.t[i] != 0.0)
        {
          IEEE_double_Accu C(*this);

          C.scale(A.t[i]);
          B.add(C);
        }

    double *tt = t;    
    t          = B.t;
    B.t        = tt;
  }

/*****************************************************************/

double IEEE_double_Accu::sum()
  {
    int    i,j;
    double a,b;

    for (i=IEEE_double_Accu_Size-1 ; (i> 0) && (t[i]==0.0) ; i--);
    for (j=i-1                     ; (j>=0) && (t[j]==0.0) ; j--);

    while (j>=0)
      {
        a = t[i] + t[j];

        if (a != t[i])
          {
            b    = t[j] - (a - t[i]);
            t[i] = 0.0;
            t[j] = 0.0;

            push(a);
            push(b);

            for (i+=3 ; (i>0) && (t[i]==0.0) ; i--);
          }
        else
          i = j;

        for (j=i-1 ; (j>=0) && (t[j]==0.0) ; j--);
      }

    for (i=IEEE_double_Accu_Size-1 ; (i>1) && (t[i]==0.0) ; i--);
    for (j=i-1                     ; (j>0) && (t[j]==0.0) ; j--);

    return t[i]+t[j];
  }

/*****************************************************************/
